/*
  This file is part of TALER
  Copyright (C) 2023-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  TALER; see the file COPYING.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file lib/exchange_api_post-management-partners.c
 * @brief functions to add a partner exchange
 * @author Christian Grothoff
 */
#include "taler/taler_json_lib.h"
#include <gnunet/gnunet_curl_lib.h>
#include <microhttpd.h>
#include "taler/exchange/post-management-partners.h"
#include "exchange_api_curl_defaults.h"
#include "taler/taler_curl_lib.h"


struct TALER_EXCHANGE_PostManagementPartnersHandle
{

  /**
   * The base URL for this request.
   */
  char *base_url;

  /**
   * The full URL for this request, set during _start.
   */
  char *url;

  /**
   * Minor context that holds body and headers.
   */
  struct TALER_CURL_PostContext post_ctx;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_EXCHANGE_PostManagementPartnersCallback cb;

  /**
   * Closure for @a cb.
   */
  TALER_EXCHANGE_POST_MANAGEMENT_PARTNERS_RESULT_CLOSURE *cb_cls;

  /**
   * Reference to the execution context.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * Offline signing key of the partner.
   */
  struct TALER_MasterPublicKeyP partner_pub;

  /**
   * Validity period start.
   */
  struct GNUNET_TIME_Timestamp start_date;

  /**
   * Validity period end.
   */
  struct GNUNET_TIME_Timestamp end_date;

  /**
   * How often will we do wad transfers to this partner.
   */
  struct GNUNET_TIME_Relative wad_frequency;

  /**
   * Wad fee to this partner.
   */
  struct TALER_Amount wad_fee;

  /**
   * Base URL of the partner exchange.
   */
  char *partner_base_url;

  /**
   * Signature.
   */
  struct TALER_MasterSignatureP master_sig;

};


/**
 * Function called when we're done processing the
 * HTTP POST /management/partners request.
 *
 * @param cls the `struct TALER_EXCHANGE_PostManagementPartnersHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response response body, NULL if not in JSON
 */
static void
handle_partners_finished (void *cls,
                          long response_code,
                          const void *response)
{
  struct TALER_EXCHANGE_PostManagementPartnersHandle *pmph = cls;
  const json_t *json = response;
  struct TALER_EXCHANGE_PostManagementPartnersResponse res = {
    .hr.http_status = (unsigned int) response_code,
    .hr.reply = json
  };

  pmph->job = NULL;
  switch (response_code)
  {
  case 0:
    /* no reply */
    res.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    res.hr.hint = "server offline?";
    break;
  case MHD_HTTP_NO_CONTENT:
    break;
  case MHD_HTTP_FORBIDDEN:
    res.hr.ec = TALER_JSON_get_error_code (json);
    res.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_CONFLICT:
    res.hr.ec = TALER_JSON_get_error_code (json);
    res.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  default:
    /* unexpected response code */
    GNUNET_break_op (0);
    res.hr.ec = TALER_JSON_get_error_code (json);
    res.hr.hint = TALER_JSON_get_error_hint (json);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d for adding exchange partner\n",
                (unsigned int) response_code,
                (int) res.hr.ec);
    break;
  }
  if (NULL != pmph->cb)
  {
    pmph->cb (pmph->cb_cls,
              &res);
    pmph->cb = NULL;
  }
  TALER_EXCHANGE_post_management_partners_cancel (pmph);
}


struct TALER_EXCHANGE_PostManagementPartnersHandle *
TALER_EXCHANGE_post_management_partners_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  const struct TALER_MasterPublicKeyP *partner_pub,
  struct GNUNET_TIME_Timestamp start_date,
  struct GNUNET_TIME_Timestamp end_date,
  struct GNUNET_TIME_Relative wad_frequency,
  const struct TALER_Amount *wad_fee,
  const char *partner_base_url,
  const struct TALER_MasterSignatureP *master_sig)
{
  struct TALER_EXCHANGE_PostManagementPartnersHandle *pmph;

  pmph = GNUNET_new (struct TALER_EXCHANGE_PostManagementPartnersHandle);
  pmph->ctx = ctx;
  pmph->base_url = GNUNET_strdup (url);
  pmph->partner_pub = *partner_pub;
  pmph->start_date = start_date;
  pmph->end_date = end_date;
  pmph->wad_frequency = wad_frequency;
  pmph->wad_fee = *wad_fee;
  pmph->partner_base_url = GNUNET_strdup (partner_base_url);
  pmph->master_sig = *master_sig;
  return pmph;
}


enum TALER_ErrorCode
TALER_EXCHANGE_post_management_partners_start (
  struct TALER_EXCHANGE_PostManagementPartnersHandle *pmph,
  TALER_EXCHANGE_PostManagementPartnersCallback cb,
  TALER_EXCHANGE_POST_MANAGEMENT_PARTNERS_RESULT_CLOSURE *cb_cls)
{
  CURL *eh;
  json_t *body;

  pmph->cb = cb;
  pmph->cb_cls = cb_cls;
  pmph->url = TALER_url_join (pmph->base_url,
                              "management/partners",
                              NULL);
  if (NULL == pmph->url)
  {
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Could not construct request URL.\n");
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  }
  body = GNUNET_JSON_PACK (
    GNUNET_JSON_pack_string ("partner_base_url",
                             pmph->partner_base_url),
    GNUNET_JSON_pack_timestamp ("start_date",
                                pmph->start_date),
    GNUNET_JSON_pack_timestamp ("end_date",
                                pmph->end_date),
    GNUNET_JSON_pack_time_rel ("wad_frequency",
                               pmph->wad_frequency),
    GNUNET_JSON_pack_data_auto ("partner_pub",
                                &pmph->partner_pub),
    GNUNET_JSON_pack_data_auto ("master_sig",
                                &pmph->master_sig),
    TALER_JSON_pack_amount ("wad_fee",
                            &pmph->wad_fee));
  eh = TALER_EXCHANGE_curl_easy_get_ (pmph->url);
  if ( (NULL == eh) ||
       (GNUNET_OK !=
        TALER_curl_easy_post (&pmph->post_ctx,
                              eh,
                              body)) )
  {
    GNUNET_break (0);
    if (NULL != eh)
      curl_easy_cleanup (eh);
    json_decref (body);
    GNUNET_free (pmph->url);
    pmph->url = NULL;
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  json_decref (body);
  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
              "Requesting URL '%s'\n",
              pmph->url);
  pmph->job = GNUNET_CURL_job_add2 (pmph->ctx,
                                    eh,
                                    pmph->post_ctx.headers,
                                    &handle_partners_finished,
                                    pmph);
  if (NULL == pmph->job)
  {
    TALER_curl_easy_post_finished (&pmph->post_ctx);
    GNUNET_free (pmph->url);
    pmph->url = NULL;
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  return TALER_EC_NONE;
}


void
TALER_EXCHANGE_post_management_partners_cancel (
  struct TALER_EXCHANGE_PostManagementPartnersHandle *pmph)
{
  if (NULL != pmph->job)
  {
    GNUNET_CURL_job_cancel (pmph->job);
    pmph->job = NULL;
  }
  TALER_curl_easy_post_finished (&pmph->post_ctx);
  GNUNET_free (pmph->partner_base_url);
  GNUNET_free (pmph->url);
  GNUNET_free (pmph->base_url);
  GNUNET_free (pmph);
}


/* end of exchange_api_post-management-partners.c */
